(*  :Title:	Ptolemy Interface  *)

(*  :Authors:	Brian Evans and James McClellan  *)

(*  :Summary:	To generate code for Berkely's Ptolemy simulation tool  *)

(*  :Context:	SignalProcessing`ObjectOriented`Ptolemy`  *)

(*  :PackageVersion:  SPP 2.9  @(#)Ptolemy.m	1.1	2/8/94  *)

(*
    :Copyright:	Copyright 1989-1994 by Brian L. Evans
		Georgia Tech Research Corporation
		All Rights Reserved

	Permission to use, copy, and distribute this software and
	its documentation is hereby granted, provided that no profit
	is made, that both the copyright notice and this permission
	notice appear in all copies, that both that copyright notice
	and this permission notice appear in supporting documentation,
	and that the name of the Georgia Tech Research Corporation
	(hereafter "GTRC"), Georgia Tech, or Georgia Institute of
	Technology not be used in advertising or publicity pertaining
	to distribution of the software without specific, written prior
	permission by the GTRC.  Permission to modify this software for
	personal use is also hereby granted provided that the modified
	version of this software is neither copied nor distributed.
	Copying or distributing modified versions of this software,
	making derivative works of this software, and marketing this
	software for profit are prohibited except with specific, written
	permission by the GTRC or by any entity to which the GTRC grants
	this authority.  GTRC makes no representations about the
	suitability of this software for any purpose.  It is provided
	"as is" without express or implied warranty.
 *)

(*  :History:	*)

(*  :Keywords:	*)

(*
    :Source:	{The Almagest: Manual for Ptolemy}, by the Dept. of
		  Electrical Engineering and Computer Science, University of
		  California at Berkeley, Berkeley, CA, 94720, c 1990.

                Evans, Brian, {A Knowledge-Based Environment for the
                  Design and Analysis of Multidimensional Multirate Signal
                  Processing Algorithms}, Ph. D. Thesis, Georgia Institute
                  of Technology, Atlanta, GA, June, 1993.
 *)

(*  :Warning:	*)

(*  :Mathematica Version:  1.2 or 2.0  *)

(*  :Limitation:  *)

(*
    :Discussion:	Ptolemy is a simulation tool for signal processing
		  algorithms.  Ptolemy is part of the Berkeley toolkit
		  for VLSI design and layout.  Thus, Ptolemy descriptions
		  can be converted into VHDL descriptions which can then
		  be post-processed and sent for fabrication.  Ptolemy also
		  has the ability to generate the equivalent C and DSP 
		  assembly language code for algorithms.
 *)

(*
    :Functions:	  BranchOperation
		  ExpressionToBranchOperations
		  FunctionToPtolemy
		  InputPortName
		  OperatorToPtolemy
		  OutputPortName
		  PlusByConstant
		  Ptolemy
		  PtolemyForm
		  PtolemyProgram
		  TimesByConstant
 *)



(*  B E G I N     P A C K A G E  *)

BeginPackage [ "SignalProcessing`ObjectOriented`Ptolemy`",
	       "SignalProcessing`Support`FilterSupport`",
	       "SignalProcessing`Support`SigProc`",
	       "SignalProcessing`Support`SupCode`" ]


If [ TrueQ[ $VersionNumber >= 2.0 ],
     Off[ General::spell ];
     Off[ General::spell1 ] ];


(*  U S A G E     I N F O R M A T I O N  *)

BranchOperation::usage =
	"BranchOperation[token, simpleOper, refs] represents one simple \
	operation simpleOper in the computation of a larger algorithm \
	with the output(s) referenced by the symbol token. \
	The larger algorithm uses the output n times. \
	BranchOperation[token, simpleOper, n, Constant] indicates that the \
	token is an alias for a constant computed according to simpleOper. \
	This object is displayed as token := simpleOper. \
	See ExpressionToBranchOperations."

ExpressionToBranchOperations::usage =
	"ExpressionToBranchOperations[expr] decomposes a nested Mathematica \
	expression (tree) into a list of simpler (branch) operations. \
	The full calling sequence is ExpressionToBranchOperations[expr, \
	pullConstants, gainflag, plusflag, a1, a2, ...]. \
	If pullConstants is True, then this routine will combine constants \
	within a product (if gainflag is True) and a summation (if incflag \
	is True) into one constant term. \
	The arguments a1, a2, ... indicate symbols in expr to be treated \
	as constants (see ConstantQ). \
	Definitions of these constants are not included as branch operations \
	because their values are unknown. \
	See BranchOperation."

FunctionToPtolemy::usage =
	"FunctionToPtolemy[head] converts a Mathematica function like Cos \
	to Ptolemy star."

InputPortName::usage =
	"InputPortName[op] indicates the name of the signal input port \
	in Ptolemy for the operator op. \
	By default, the function returns \"input\"."

OperatorToPtolemy::usage =
	"OperatorToPtolemy[op[parameters]] converts an operator to a \
	Ptolemy star."

OutputPortName::usage =
	"OutputPortName[op] indicates the name of the signal output port \
	in Ptolemy for the operator op. \
	By default, the function returns \"output\"."

PlusByConstant::usage =
	"PlusByConstant[a, const] indicates the sum of a and the \
	constant const."

Ptolemy::usage =
	"Ptolemy is a block diagram simulation environment developed at \
	the Department of Electrical Engineering and Computer Science \
	at the University of California at Berkeley. \
	It features the ability to simulate algorithms that mix \
	synchronous data flow, dynamic data flow, and discrete event \
	methods of computation. \
	The environment can also generate C and assembly language programs. \
	See also PtolemyProgram and PtolemyForm."

PtolemyForm::usage =
	"PtolemyForm[expr] generates the Ptolemy code that is equivalent \
	to the signal processing expression expr. \
	It is called by PtolemyProgram to generate self-contained \
	Ptolemy simulations. \
	See also Ptolemy and PtolemyProgram."

PtolemyProgram::usage =
	"PtolemyProgram[expr, {v1, v1min, v1max}, {v2, v2min, v2max}, ...] \
	generates a complete Ptolemy program for simulation based \
	on the processing of in variables v1, v2, ... over the specified \
	range of values. \
	The following options are supported: \
	Dialogue (see Dialogue), Prolog (see $PtolemyProlog), and \
	Plot (True or Automatic will plot the result of the algorithm expr). \
	For example, the user can add aa -> 2 to specify that the constant \
	aa in expr be assigned the value 2 in the generated Ptolemy code. \
	See also Ptolemy and PtolemyForm."

$PtolemyProlog::usage =
	"$PtolemyProlog is a global variable containing the definitions \
	to make the generated Ptolemy code work properly. \
	The default value will find the file \"headerX.X.pt\" on \
	Mathematica's path, read it in, and write it out, \
	where X.X is $PtolemyVersion. \
	This is useful if you are running the Ptolemy interpreter on \
	another file system. \
	Otherwise, you can set this value to \
	StringJoin[ \"(load \\\"\", \
	FindFile[ \"SignalProcessing`ObjectOriented`headerX.X.pt\", $Path ], \
	\")\\\"\" ], where X.X is the version of Ptolemy you are using, \
	which will cause Ptolemy to load the header as the first step in \
	running the simulation."

$PtolemyVersion::usage =
	"$PtolemyVersion indicates the current version of Ptolemy being used. \
	Currently, only version 0.3 of Ptolemy is supported."

TimesByConstant::usage =
	"TimesByConstant[a, const] indicates the product of a and the \
	constant const."

(*  E N D     U S A G E     I N F O R M A T I O N  *)


Begin[ "`Private`" ]


(*  M E S S A G E S  *)

(*  usage messages  *)

Ptolemy::command1 = "!interpreter <file>"
Ptolemy::command2 = "PtolemyProgram[ ... ] >> \"!interpreter\""

Ptolemy::directions1 =
	"If Ptolemy is installed on your machine, you can test your \
	simulation by writing the output to a file (it is common to add a \
	.pt extension for Ptolemy files), then by editing the file if \
	necessary, and finally by evaluating"

Ptolemy::directions2 =
	"Or, you can send the output of this command directly to the \
	Ptolemy simulator:"

Ptolemy::mma =
	"In order to generate Ptolemy code, the Mathematica functions \
	Continuation and LineBreak have been altered. \
	To return them to their default values, evaluate \
	Clear[Continuation, LineBreak]."

(*  message for conversion from SPP/Mma notation to Ptolemy  *)

Ptolemy::Abs = "(star `1` FloatAbs)"
Ptolemy::Ceiling = "(star `1` Ceiling)"
Ptolemy::Cos = "(star `1` Cos)"
Ptolemy::CPulse = "(star `1` FloatRect)\n(setstate `1` width `2`)"
			(* not using height state built into FloatRect *)
Ptolemy::CStep = "(star `1` Step)"
Ptolemy::Dirichlet =
    "(star `1` Dirichlet)\n(setstate `1` length `2`)"
Ptolemy::DC = "(star `1` FloatDC)\n(setstate `1` level \"`1`\")"
Ptolemy::Downsample = "(star `1` DownSample)\n(setstate `1` factor `2`)"
Ptolemy::Exp = "(star `1` Exp)"
Ptolemy::FIR =
     "(star `1` FIR)\n(setstate `1` taps `2`)"
Ptolemy::Floor = "(star `1` Floor)"
Ptolemy::Gain = "(star `1` FloatGain)\n(setstate `1` gain \"`2`\")"
Ptolemy::Impulse = "(star `1` Impulse)"
Ptolemy::IntegerPower = "(star `1` IntegerPower`2`)"
Ptolemy::NegIntegerPower = "(star `1` NegIntegerPower`2`)"
Ptolemy::Log = "(star `1` Log)"
Ptolemy::Plus = "(star `1` FloatSum)"
Ptolemy::PolyphaseDownsample =
    "(star `1` FIR)\n\
    (setstate `1` decimation `2`)\n\
    (setstate `1` taps \"`3`\")"
Ptolemy::PolyphaseUpsample =
    "(star `1` FIR)\n\
    (setstate `1` interpolation `2`)\n\
    (setstate `1` taps \"`3`\")"
Ptolemy::Power = "(star `1` Power)"
Ptolemy::PowerBase = "(star `1` PowerKnownBase)\n(setstate `1` base `2`)"
Ptolemy::Pulse = "(star `1` FloatRect)\n(setstate `1` width `2`)"
			(* not using height state built into FloatRect *)
Ptolemy::Reciprocal = "(star `1` Reciprocal)"
Ptolemy::Rev = "(star `1` Reverse)"
Ptolemy::Shift = "(star `1` Shift)\n(setstate `1` shift `2`)"
Ptolemy::Sign = "(star `1` Sgn)"
Ptolemy::Sin = "(star `1` Sin)"
Ptolemy::Sinc = "(star `1` Sinc)"
Ptolemy::Step = "(star `1` Step)"
Ptolemy::PolyphaseResample =
    "(star `1` FIR)\n\
    (setstate `1` interpolation `2`)\n\
    (setstate `1` taps \"`3`\")\n\
    (setstate `1` decimation `4`)"
Ptolemy::Sqrt = "(star `1` Sqrt)"
Ptolemy::Times = "(star `1` FloatProduct)"
Ptolemy::Upsample = "(star `1` UpSample)\n(setstate `1` factor `2`)"

(*  Other strings necessary for code generation  *)

Ptolemy::algheader =
	"\n#\n#\tAlgorithm\n#"
Ptolemy::comment =
	"#\tCannot translate function `` for token ``\n"
Ptolemy::connect =
	"\n(nodeconnect (`1` `2`) `3`)"
Ptolemy::constant =
	"# `3`\n(state `1` float \"`2`\")"
Ptolemy::constheader =
	"\n#\n#\tConstants\n#"
Ptolemy::iteration =
	"(star `1` FloatRamp)\n(setstate `1` value \"`2`\")\n\
	(node `1`)\n(nodeconnect (`1` output) `1`)"
Ptolemy::iteration1 =
	"(star `1` FloatRamp)\n(setstate `1` value \"`2`\")\n\
	(setstate `1` step `3`)\n(node `1`)\n\
	(nodeconnect (`1` output) `1`)"
Ptolemy::mconnect =
	"\n(nodeconnect (`1` \"`2`#`3`\") `4`)"
Ptolemy::node =
	"\n(node `1`)\n(nodeconnect (`1` `2`) `1`)"
Ptolemy::numports =
	"(numports `1` `2` `3`)"
Ptolemy::prolog =
	"The prolog option did not evaluate to a string. \
	A null string will be used as the header."
Ptolemy::simheader =
	"\n#\n#\tRun Simulation\n#"
Ptolemy::terminate =
	"(run `1`)\n(wrapup)"
Ptolemy::unknown =
	"Do not know how to translate the Mathematica function `` \
	into Ptolemy form."
Ptolemy::varheader =
	"\n#\n#\tVariables\n#"
Ptolemy::xgraph =
	"(star `1` Xgraph)"
Ptolemy::xygraph =
	"(star `1` XYgraph)\n(nodeconnect (`1` xInput) `2`)"

(*  mimicking the power function in Ptolemy  *)

Ptolemy::intPowerGalaxyBegin =
"(defgalaxy IntegerPower`1`\n\
\t(domain SDF)\n\
\t(descriptor \"Power function takes input to a power of `1`\")\n\
\t(star fork Fork)\n\
\t(star times FloatProduct)\n\
\t(numports fork output `1`)\n\
\t(numports times input `1`)\n"

Ptolemy::intPowerGalaxyConnect =
	"\t(connect (fork \"output#`1`\") (times \"input#`1`\"))"

Ptolemy::intPowerGalaxyEnd =
"\n\
\t(input input (fork input))\n\
\t(output output (times output))\n\
)\n\n"

Ptolemy::intNegPowerGalaxy =
"(defgalaxy IntegerPowerNeg`1`\n\
\t(domain SDF)\n\
\t(descriptor \"Power function takes input to a power of `1`\")\n\
\t(star reciprocal Reciprocal)\n\
\t(star intPower IntegerPower`1`)\n\
\n\
\t(connect (reciprocal output) (intPower input))\n\
\n\
\t(input input (reciprocal input))\n\
\t(output output (intPower output))\n\
)"


(*  L O C A L     V A R I A B L E S  *)

constantTokenList = {}
dcTokenList = {}


(*  G L O B A L     V A R I A B L E S  *)

(*  BranchOperation data structure  *)
Format[ BranchOperation[name_, op_, ___] ] := SequenceForm[ name, " := " , op ]

(*  $PtolemyVersion  *)
$PtolemyVersion = 0.3

(*  $PtolemyProlog is generated in Unix by the "build_header" script      *)
$PtolemyProlog :=
  ReadTextFile[
    FindFile[ StringJoin[ "SignalProcessing`ObjectOriented`header",
			  ToString[$PtolemyVersion],
			  ".pt" ],
	      $Path ] ]


(*  L O C A L     H A S H     T A B L E     R O U T I N E S  *)

If [ ! TrueQ[ $VersionNumber >= 2.0 ],
     ToCharacterCode[ str_ ] := Map[ ToASCII, Characters[str] ] ]

hashTableLength = 0
hashTable = {}
assocTable = {}

hashCoeff[ n_ ] := {1, 3, 5, 2}[[ Mod[n, 4] + 1 ]]

hashCode[ expr_ ] := 
Block [	{chars, entry, factors, len},
	entry = ToString[ InputForm[ expr ] ];
	chars = ToCharacterCode[entry] - 26;
	len = Length[chars];
	factors = Table[ hashCoeff[i], {i, 1, len} ];
	Mod[factors . chars, hashTableLength] + 1 ]

hashTableInit[len_] :=
Block [	{},
	hashTableLength = len;
	hashTable = Table[ {}, {len} ];
	assocTable = Table[ {}, {len} ];
	Null ];

hashTableSearch[ expr_ ] :=
Block [	{index, pattern, possibleMatches},
	index = hashCode[expr];
	possibleMatches = hashTable[[ index ]];
	pos = Position[ possibleMatches, expr ];
	If [ SameQ[pos, {}],
	     { False, Null },
	     { True, assocTable[[ index, pos[[1,1]] ]] } ] ]

hashTableAdd[ expr_, key_ ] :=
Block [	{entry},
	entry = hashCode[expr];
	AppendTo[ hashTable[[entry]], expr ];
	AppendTo[ assocTable[[entry]], key ];
	Null ]

hashTableDestroy[ ] := Block [ {}, hashTable = assocTable = {}; Null ]


(*  O T H E R     S U P P O R T I N G     F U N C T I O N S  *)

(*  Backwards compatibility with Mathematica 1.2  *)
If [ ! TrueQ[ $VersionNumber >= 2.0 ],
     tolower[ char_ ] :=
     	If [ 65 <= ToASCII[char] <= 90, FromASCII[32 + ToASCII[char]], char ];
     ToLowerCase[string_] :=
	Apply[StringJoin, Map[ tolower, Characters[string] ] ] ]

(*  addOutput  *)
addOutput[ BranchOperation[ token_, expr_, num_, rest___ ] ] :=
	BranchOperation[token, expr, num+1, rest]

(*  addToProlog  *)
makePowerGalaxy[n_] :=
Block [ {i},
	ptolemyStringJoin[
		Join[ { ptolemyString[ Ptolemy::intPowerGalaxyBegin, n ] },
		      Table[ ptolemyString[Ptolemy::intPowerGalaxyConnect, i ],
			     {i, 1, n} ],
		      { Ptolemy::intPowerGalaxyEnd } ],
		addOneNewLine ] ] /;
	n > 0

makePowerGalaxy[n_] :=
	StringJoin[ makePowerGalaxy[-n],
		    ptolemyString[ Ptolemy::NegIntegerPower, -n ] ] /;
	n < 0

addToProlog[ expr_ ] :=
	Block [	{code = "", powers = {}},
		expr //.
		      ( Power[len_, n_Integer] :>
			Block [ {},
				AppendTo[powers, n];
				If [ n < 0, AppendTo[powers, -n] ];
				code = StringJoin[code, makePowerGalaxy[n]];
				len ] /;
			( n > 1 || n < -1 ) && ! MemberQ[powers, n] );
		code ]

(*  constant2node  *)
constant2node[ token_ ] :=
	StringJoin[ ptolemyString[Ptolemy::DC, token], "\n", makeNode[token] ]

(*  connectInputs  *)
singleConnect[ token_, inport_, outport_, a_ ] :=
Block [	{},
	StringJoin[ makeNode[token, outport],
		    "\n",
		    ptolemyString[ Ptolemy::connect, token, inport, a ] ] ]

multipleConnect[ token_, inport_, outport_, alist_ ] :=
Block [	{code = "", defstr, i, len},
	len = Length[alist];
	code = Table[ ptolemyString[ Ptolemy::mconnect,
				     token, inport, i, alist[[i]] ],
		      {i, 1, len} ];
	defstr =
	    StringJoin[ makeNode[token, outport],
			"\n",
			ptolemyString[ Ptolemy::numports, token, inport, len ],
			"\n" ];
	Apply[ StringJoin, Prepend[ code, defstr ] ] ]

connectInputs[ token_, op_, a_ ] :=
	singleConnect[ token, InputPortName[op], OutputPortName[op], a ]
connectInputs[ token_, op_, a1_, a2_, arest___ ] :=
	multipleConnect[ token, InputPortName[op],
			 OutputPortName[op], {a1, a2, arest} ]

(*  constantTokenQ  *)
constantTokenQ[x_] := MemberQ[ constantTokenList, x ]

(*  decomposeExpression calls the recursive function bottomUpRewrite  *)
flattenedForm = {}
flattenedIndex = 0

makeConst[c_] := StringJoin[ ToString[Head[c]], "const" ]

extractConstants[ e_?AtomQ, ___ ] := e
extractConstants[ e_, _, False, False ] := e
extractConstants[ e_, constlist_, timesByflag_, plusByflag_ ] :=
Block [	{},
	constantQ[x_] := ConstantQ[x, constlist];
	notConstantQ[x_] := ! constantQ[x];
	split[h_, head_, byconsthead_] :=
		Block [ {b, c, constToken, factors},
			factors = Apply[List, h];
			c = Apply[ head, Select[factors, constantQ] ];
			b = Apply[ head, Select[factors, notConstantQ] ];
			constToken =
			    If [ MemberQ[constlist, c],
				 c,
				 rewriteNode[c, makeConst[c], 1, Constant] ];
			byconsthead[ extractConstants[b, constlist,
						      timesByflag, plusByflag],
				     constToken ] ];
	e /.
	    { h_Times :>
		split[h, Times, TimesByConstant] /;
		timesByflag && Apply[ Or, Map[constantQ, Apply[List, h]] ],
	      h_Plus :>
		split[h, Plus, PlusByConstant] /;
		plusByflag && Apply[ Or, Map[constantQ, Apply[List, h]] ] } ]

rewriteNode[ h_[p___][a___] ] := rewriteNode[ h[p][a], h, 1 ]
rewriteNode[ h_[a___] ] := rewriteNode[ h[a], h, 1 ]
rewriteNode[ x_ ] := x

rewriteNode[ expr_, head_, branchinfo___ ] :=
	Block [	{found, index, token, value},
		{found, value} = hashTableSearch[ expr ];
		If [ found,
		     {token, index} = value;
		     flattenedForm[[index]] =
			 addOutput[ flattenedForm[[index]] ],
		     token = Unique[ ToLowerCase[ ToString[head] ] ];
		     AppendTo[ flattenedForm,
			       BranchOperation[token, expr, branchinfo] ];
		     flattenedIndex++;
		     hashTableAdd[ expr, {token, flattenedIndex} ] ];
		token ]

bottomUpRewrite[ x_ ] := rewriteNode[x] /; flatExprQ[x]
bottomUpRewrite[ x_ ] := rewriteNode[ Map[ bottomUpRewrite, x ] ]

decomposeExpression[ expr_ ] := 
	decomposeExpression[ expr, False ]
decomposeExpression[ expr_, False ] :=
	decomposeExpression[ expr, False, False, False ]

decomposeExpression[ e_, pullconstants_, gainflag_, incflag_, constants___ ] :=
	Block [	{constlist, expr, hashLength},

		(*  hash table size depends on size of expr  *)
		hashLength = nearestPrime[ Ceiling[LeafCount[expr] / 5] ];
		hashTableInit[ Max[hashLength, 5] ];

		flattenedForm = {};
		flattenedIndex = 0;
		constlist = If [ TrueQ[pullconstants], ToList[constants], {} ];
		expr = If [ TrueQ[pullconstants],
			    extractConstants[e, constlist, gainflag, incflag],
			    e ];
		bottomUpRewrite[expr];
		hashTableDestroy[];
		flattenedForm ]

(*  flatExprQ  *)
flatExprQ[ h_[p___][a1_, a___] ] := Apply[ And, Map[AtomQ, {a1, a}] ]
flatExprQ[ h_[a1_, a___] ] := Apply[ And, Map[AtomQ, {a1, a}] ]

(*  generateConstants  *)
generateConstants[rules_][sym_Symbol] :=
Block [	{str, value},
	value = sym /. rules;
	If [ SameQ[value, sym],
	     value = 1;
	     str = "# Please fill in a value besides 1 for this constant",
	     str = Short[InputForm[value]] ];
	If [ ! NumberQ[value], value = N[value, 16] ];
	ptolemyString[ Ptolemy::constant, sym, value, str ] ]

(*  generateIteration  *)
generateIteration[ { v_, vmin_, vmax_ } ] :=
	ptolemyString[ Ptolemy::iteration, v, vmin ]
generateIteration[ { v_, vmin_, vmax_, vinc_ } ] :=
	ptolemyString[ Ptolemy::iteration1, v, vmin, vinc ]

(*  generatePair -- coded in the order of most frequent use  *)
generatePair[ format_, token_Symbol, rest___ ] :=
	ptolemyString[ format, token, rest ]
generatePair[ format_, token_String, rest___ ] :=
	generatePair[ format, Unique[tokenStr], rest ]

(*  iterationPoints  *)
iterationPoints[ { v_, vmin_, vmax_, vinc_:1 } ] := (1 + (vmax - vmin) / vinc)

(*  makeNode  *)
makeNode[ token_, outport_:"output" ] :=
	ptolemyString[ Ptolemy::node, token, outport ]

(*  makePlotter  *)
makePlotter[ string_, fun_, rest___ ] :=
Block [	{message, token},
	token = Unique[string];
	message = MessageName[Ptolemy, string];
	ptolemyStringJoin[ { ptolemyString[ message, token, rest],
			     ptolemyString[ Ptolemy::connect,
					    token, "input", fun ] },
			   addOneNewLine ] ]

(*  nearestPrime  *)
nearestPrime[ x_ ] := Block [ {i = x}, While [ ! PrimeQ[i], i-- ]; i ]

(*  powerGalaxy  *)
powerGalaxy[ token_Symbol, base_Symbol, i_Integer ] :=
	StringJoin[ generatePair[ Ptolemy::IntegerPower, token, i ],
		    connectInputs[ token, Power, base ] ] /;
	(i > 1) || (i < -1)
powerGalaxy[ token_Symbol, base_Symbol, 1/2 ] :=
	StringJoin[ generatePair[ Ptolemy::Sqrt, token ],
		    connectInputs[ token, Sqrt, base ] ]
powerGalaxy[ token_Symbol, base_Symbol, -1 ] :=
	StringJoin[ generatePair[ Ptolemy::Reciprocal, token ],
		    connectInputs[ token, Power, base ] ]
powerGalaxy[ token_, E, exp_ ] :=
	StringJoin[ FunctionToPtolemy[ Exp, "Exp", token ],
		    connectInputs[ token, Exp, exp ] ]
powerGalaxy[ token_, base_?NumberQ, exp_ ] :=
	StringJoin[ generatePair[ Ptolemy::PowerBase, token, N[Log[base], 16] ],
		    connectInputs[ token, Power, exp ] ]
powerGalaxy[ token_, base_, exp_ ] :=
	StringJoin[ FunctionToPtolemy[ Power, "Power", token ],
		    connectInputs[ token, Power, base, exp ] ]

(*  preprocess  *)
preprocess[ x_ ] :=
Block [	{},
	x //. ( Convolve[n_][args1___, h_[hargs__], args2___] :>
		  h[hargs][ Convolve[n][args1, args2] ] /;
		  SameQ[h, FIR] || SameQ[h, IIR] ) ]

(*  ptolemyString  *)
(*  using TeXForm is the only way to make newlines display as newlines *)
ptolemyString[ s_String ] := s
ptolemyString[ str_, args__ ] :=
	ToString[
	  TeXForm[
	    Apply[ StringForm, Prepend[ Map[ToString, {args}], str ] ] ] ]

(*  ptolemyPlotter  *)
ptolemyPlotter[ fun_ ] :=
	makePlotter[ "xgraph", fun ]
ptolemyPlotter[ fun_, var_ ] :=
	makePlotter[ "xygraph", fun, var ]

(*  ptolemyStringJoin  *)
addOneNewLine[ s_String ] := StringJoin[ "\n", s ]
addTwoNewLines[ s_String ] := StringJoin[ "\n\n", s ]
ptolemyStringJoin[ {}, ___ ] := ""
ptolemyStringJoin[ {x_}, ___ ] := x
ptolemyStringJoin[ x_List, fun_:addTwoNewLines ] := 
	Apply[ StringJoin, Prepend[ Map[fun, Rest[x]], First[x] ] ]

(*  ptolemyVector  *)
toStringWithSpace[ 0. ] := StringJoin[ "0.0 " ]
toStringWithSpace[ x_ ] := StringJoin[ ToString[x], " " ]
ptolemyVector[ vec_List ] :=
	StringJoin[ "\"",
		    Apply[ StringJoin, Map[ toStringWithSpace, vec ] ],
		    "\"" ]

ptolemyVector[ Hold[ReadList[file_, type_]] ] :=
	StringJoin[ "\"<", ToString[file], "\"" ]

(*  simulationTrailer  *)
simulationTrailer[num_, plot_, fun_, {var_}] :=
Block [	{pp, trailer},
	trailer = ptolemyString[ Ptolemy::terminate, Ceiling[numpoints] ];
	Which [ TrueQ[plot],
		  pp = If [ SameQ[Head[var], Symbol],
		  	    ptolemyPlotter[fun, var],
		  	    ptolemyPlotter[fun] ];
		  ptolemyStringJoin[ { pp, trailer } ],
		SameQ[plot, False],
		  trailer,
		True,
		  ptolemyStringJoin[ { ToString[plot], trailer } ] ] ]

(*  validConstantQ  *)
validConstantQ[ x_ ] :=
	If [ SameQ[ Context[x], "System`" ], ConstantQ[x], True ]


(*  E X P O R T E D     F U N C T I O N S  *)

(*  ExpressionToBranchOperations  *)
ExpressionToBranchOperations[ expr_, rest___ ] :=
	decomposeExpression[ expr, rest ]

(*  FunctionToPtolemy  *)
FunctionToPtolemy[ h_Symbol ] :=
	FunctionToPtolemy[ h, ToString[h] ]
FunctionToPtolemy[ h_Symbol, hstr_ ] :=
	FunctionToPtolemy[ h, hstr, Unique[ToLowerCase[hstr]] ]
FunctionToPtolemy[ h_Symbol, hstr_, token_ ] :=
	Check[ generatePair[ MessageName[Ptolemy, hstr], token ],
	       MyMessage[ Ptolemy::unknown,
			  ToString[ StringForm[ Ptolemy::comment, h, token ] ],
			  h ] ]

(*  InputPortName  *)
InputPortName[x_] := "input"

(*  OperatorToPtolemy-- connect the Ptolemy objects later *)
OperatorToPtolemy[ op_Symbol[p___] ] :=
	OperatorToPtolemy[ op[p], ToLowerCase[ToString[op]] ]

OperatorToPtolemy[ Downsample[m_, n_, ___], token_ ] :=
	generatePair[ Ptolemy::Downsample, token, m ]
OperatorToPtolemy[ FIR[n_, coeffs_, ___], token_ ] :=
	generatePair[ Ptolemy::FIR, token, ptolemyVector[coeffs] ]
OperatorToPtolemy[ PolyphaseDownsample[m_, FIR[n_, coeffs_], n_, ___],
		   token_ ] :=
	generatePair[ Ptolemy::PolyphaseDownsample, token,
		      m, ptolemyVector[coeffs] ]
OperatorToPtolemy[ PolyphaseResample[l_, FIR[n_, coeffs_], m_, n_, ___],
		   token_ ] :=
	generatePair[ Ptolemy::PolyphaseResample, token,
		      l, ptolemyVector[coeffs], m ]
OperatorToPtolemy[ PolyphaseUpsample[l_, FIR[n_, coeffs_], n_, ___],
		   token_ ] :=
	generatePair[ Ptolemy::PolyphaseDownsample, token,
		      l, ptolemyVector[coeffs] ]
OperatorToPtolemy[ Shift[l_, n_, ___], token_ ] :=
	generatePair[ Ptolemy::Shift, token, l ]
OperatorToPtolemy[ Upsample[l_, n_, ___], token_ ] :=
	generatePair[ Ptolemy::Upsample, token, l ]

(*  OutputPortName  *)
OutputPortName[x_] := "output"

(* PlusByConstant *)
Format[ PlusByConstant[input_, inc_] ] := Plus[input, inc]

(*  PtolemyForm  *)

PtolemyForm[ BranchOperation[ token_, c_, num_, Constant, ___ ] ] :=
	Block [ {},
		AppendTo[ constantTokenList, token ];
		generatePair[ Ptolemy::constant,
			      token, N[c, 16], Short[InputForm[c]] ] ]

PtolemyForm[ BranchOperation[ token_, Power[len_, n_], ___ ] ] :=
	powerGalaxy[token, len, n]

PtolemyForm[ BranchOperation[ token_, h_Symbol[len_, n_], ___ ] ] :=
	StringJoin[ generatePair[ MessageName[Ptolemy, ToString[h]],
				  token, len ],
		    connectInputs[ token, h, n ] ] /;
	MemberQ[ {CPulse, Dirichlet, Pulse}, h ]

PtolemyForm[ BranchOperation[ token_, PlusByConstant[input_, inc_], ___ ] ] :=
	Block [	{code = ""},
		If [ ! MemberQ[dcTokenList, inc],
		     code = StringJoin[constant2node[inc], "\n"];
		     AppendTo[dcTokenList, inc] ];
		StringJoin[ code,
			    generatePair[ Ptolemy::Plus, token ],
			    connectInputs[ token, PlusByConstant,
					   input, inc ] ] ]

PtolemyForm[ BranchOperation[ token_, TimesByConstant[input_, gain_], ___ ] ] :=
	StringJoin[ generatePair[ Ptolemy::Gain, token, gain ],
		    connectInputs[ token, TimesByConstant, input ] ]

PtolemyForm[ BranchOperation[ token_, h_Symbol[a___], ___ ] ] :=
	StringJoin[ FunctionToPtolemy[ h, ToString[h], token ],
		    connectInputs[ token, h, a ] ]

PtolemyForm[ BranchOperation[ token_, op_Symbol[p__][a___], ___ ] ] :=
	StringJoin[ OperatorToPtolemy[ op[p], token ],
		    connectInputs[ token, op, a ] ]

(* PtolemyProgram *)
Options[ PtolemyProgram ] :=
	{ Dialogue -> False, Prolog :> $PtolemyProlog, Plot -> True  }

PtolemyProgram[ expr_, a___, {v1_Symbol, v1min_, v1max_, i___} ] :=
	ptolemyProgram[ expr, {a, {v1, v1min, v1max, i}}, {} ]

PtolemyProgram[ expr_, a___, {v1_Symbol, v1min_, v1max_, i1___}, op1_, o___ ] :=
	ptolemyProgram[ expr, {a, {v1, v1min, v1max, i1}}, ToList[op1, o] ] /;
	! MatchQ[ op1, {v2_Symbol, v2min_, v2max_, i2___} ]

ptolemyProgram[ e_, rangelist_, oplist_ ] :=
	Block [	{algcode, branchform, code, constcode, constlist,
		 dialogue, endcode, expr, moreCode, numpoints,
		 options, plot, prolog, rules, varcode, varlist},

		constantTokenList = dcTokenList = {};

		options = oplist ~Join~ Options[PtolemyProgram];
		prolog = Replace[ Prolog, options ];
		dialogue = Replace[ Dialogue, options ];
		plot = Replace[ Plot, options ];
		rules = RemoveOptions[ options,
				       Map[ First, Options[PtolemyProgram] ] ];
		If [ TrueQ[ $VersionNumber >= 2.0 ], rules = Dispatch[rules] ];

		expr = preprocess[e];

		varlist = First[ Transpose[rangelist] ];
		constlist = Complement[ GetVariables[expr], varlist ];
		constlist = Select[ constlist, validConstantQ ];

		If [ InformUserQ[dialogue],
		     Print[ "The list of variables is ", varlist, "." ];
		     Print[ "The list of constants is ", constlist, "." ];
		     If [ ! EmptyQ[ constlist ],
			  Print[ "You must give the constants values ",
				 "in the generated Ptolemy code ",
				 "before you run the simulation.  Each ",
				 "constant currently has a value of 1." ] ] ];

		If [ ! SameQ[ Head[prolog], String],
		     Message[ Ptolemy::prolog ]; prolog = "" ];

		moreCode = addToProlog[expr];
		If [ ! SameQ[modeCode, ""],
		     prolog = ptolemyStringJoin[ {prolog, moreCode} ] ];

		If [ SameQ[plot, Automatic], plot = True ];

		(* Generate constants as states and variables as stars *)
		constcode = ptolemyStringJoin[ Map[ generateConstants[rules],
						    constlist ] ];
		varcode = ptolemyStringJoin[ Map[ generateIteration,
						  rangelist ] ];

		(* Generate the algorithm first by converting the algorithm *)
		(* to branch operations and then by converting each branch  *)
		(* operation to Ptolemy form *)
		branchform = ExpressionToBranchOperations[expr, True, True,
							  True, constlist];
		If [ SameQ[dialogue, All],
		     Print["Here is the algorithm decomposed into branches:"];
		     Print[ ColumnForm[branchform] ] ];
		algcode =
			ptolemyStringJoin[ Map[PtolemyForm, branchform] ];

		(* Generate the instructions to invoke the simulation *)
		numpoints = Apply[Times, Map[iterationPoints, rangelist] ];
		endcode = simulationTrailer[Ceiling[numpoints], plot,
					    First[Last[branchform]], varlist];

		(* Combine the different pieces into a complete program *)
		code = ptolemyStringJoin[ 
			{ prolog,
			  Ptolemy::constheader, constcode,
			  Ptolemy::varheader, varcode,
			  Ptolemy::algheader, algcode,
			  Ptolemy::simheader, endcode } ];

		(* Cleanup and direct the user *)
		constantTokenList = dcTokenList = {};

		If [ InformUserQ[dialogue],
		     Print[ Ptolemy::directions1 ];
		     Print[ Ptolemy::command1 ];
		     Print[ Ptolemy::directions2 ];
		     Print[ Ptolemy::command2 ] ];

		Message[ Ptolemy::mma ];

		(* Return the code *)
		Format[ Continuation[n_] ] := "";
		Format[ LineBreak[n_] ] := "";
		OutputForm[ code ] ]

(* TimesByConstant *)
Format[ TimesByConstant[input_, gain_] ] := Times[gain, input]


(*  E N D     P A C K A G E  *)

End[]
EndPackage[]

If [ TrueQ[ $VersionNumber >= 2.0 ],
     On[ General::spell ];
     On[ General::spell1 ] ];


(*  C H A N G E S  *)

Unprotect[FIR];
FIR/: InputPortName[FIR] := "signalIn"
FIR/: OutputPortName[FIR] := "signalOut"
Protect[FIR];


(*  H E L P     I N F O R M A T I O N  *)

Protect[ "$PtolemyProlog", "$PtolemyVersion" ]

Block [ {newfuns},
	newfuns =
	{ BranchOperation,	ExpressionToBranchOperations,
	  FunctionToPtolemy,	InputPortName,
	  OperatorToPtolemy,	OutputPortName,
	  PlusByConstant,	Ptolemy,
	  PtolemyForm,		PtolemyProgram,
	  TimesByConstant };
	SPfunctions = Combine[SPfunctions, newfuns];
	Apply[Protect, newfuns] ]


(*  E N D I N G     M E S S A G E  *)

Print[ "Ptolemy code generation module loaded." ]
Null
