(*  :Title:	Supplemental for Multidimensional Convolution  *)

(*  :Authors:	Brian Evans, James McClellan  *)

(*  :Summary:	Continuous/Discrete convolution for multidimensional signals *)

(*  :Context:	SignalProcessing`Support`Convolution`  *)

(*  :PackageVersion:  SPP 2.9  @(#)MDConvolution.m	1.1	2/8/94  *)

(*
    :Copyright:	Copyright 1989-1994 by Brian L. Evans
		Georgia Tech Research Corporation
		All Rights Reserved

	Permission to use, copy, and distribute this software and
	its documentation is hereby granted, provided that no profit
	is made, that both the copyright notice and this permission
	notice appear in all copies, that both that copyright notice
	and this permission notice appear in supporting documentation,
	and that the name of the Georgia Tech Research Corporation
	(hereafter "GTRC"), Georgia Tech, or Georgia Institute of
	Technology not be used in advertising or publicity pertaining
	to distribution of the software without specific, written prior
	permission by the GTRC.  Permission to modify this software for
	personal use is also hereby granted provided that the modified
	version of this software is neither copied nor distributed.
	Copying or distributing modified versions of this software,
	making derivative works of this software, and marketing this
	software for profit are prohibited except with specific, written
	permission by the GTRC or by any entity to which the GTRC grants
	this authority.  GTRC makes no representations about the
	suitability of this software for any purpose.  It is provided
	"as is" without express or implied warranty.
 *)

(*  :History:	*)

(*  :Keywords:	convolution  *)

(*
    :Source:	B. L. Evans, "A Knowledge-Based Environment for the Design
		  and Analysis of Multidimensional Multirate Signal Processing
		  Algorithms", Ph. D. Thesis, Georgia Institute of Technology,
		  Atlanta, GA, June, 1993.
 *)

(*  :Warning:	*)

(*  :Mathematica Version:  1.2 or 2.0  *)

(*  :Limitation:  *)

(*  :Discussion:  *)

(*  :Functions:	  *)


If [ TrueQ[ $VersionNumber >= 2.0 ],
     Off[ General::spell ];
     Off[ General::spell1 ] ]


(*  B E G I N     P A C K A G E  *)

BeginPackage[ "SignalProcessing`Support`MDConvolution`" ]
EndPackage[]

BeginPackage[ "SignalProcessing`Support`Convolution`",
	      "SignalProcessing`Support`SigProc`",
	      "SignalProcessing`Support`Lisp`",
	      "SignalProcessing`Support`SupCode`" ]


Begin["`Private`"]


(*  mDconvolve  *)
mDconvolveTwoIntervals[pair_List, tlist_, time_] :=
	Block [	{dimension, f, fp, g, gp, i, taulist},
		dimension = Length[tlist];
		taulist = Table[ tau[i], {i, 1, dimension} ];
		f = pair[[1]];
		g = pair[[2]];
		fp = First[f] /. ReplaceWith[tlist, taulist];
		gp = First[g] /. ReplaceWith[tlist, tlist - taulist];
		mDconvolve[ fp gp, tlist, f[[2]], f[[3]],
			    g[[2]], g[[3]], time, taulist ] ]

mDconvolve[ result_, {}, lf_, uf_, lg_, ug_, time_, taulist_ ] := result

mDconvolve[ Fintervals_List, tlist_, lf_, uf_, lg_, ug_, time_, taulist_ ] :=
	Block [	{conv, finterval, i, leftendpoint, numintervals,
		 result, rightendpoint, thefun},
		numintervals = Length[Fintervals];
		result = {};
		For [ i = 1, i <= numintervals, i++,
		      finterval = Fintervals[[i]];
		      thefun = finterval[[1]];
		      leftendpoint = finterval[[2]];
		      rightendpoint = finterval[[3]];
		      conv = mDconvolve[ thefun, tlist, lf, uf,
					 lg, ug, time, taulist ];
		      conv = Map[ extendFinterval[leftendpoint, rightendpoint],
				  conv ];
		      result = Join[ result, conv ] ];
		result ]

mDconvolve[ fun_, tlist_, lf_, uf_, lg_, ug_, time_, taulist_ ] :=
	mDconvolve[ mDconvolveTwoIntervals[ {fun, First[lf], First[uf]},
					    {  1, First[lg], First[ug] },
					    First[tlist], time,
					    First[taulist] ],
		    Rest[tlist], Rest[lf], Rest[uf], Rest[lg], Rest[ug],
		    time, Rest[taulist] ]


extendFinterval[ left_, right_ ] [ { } ] := { }

extendFinterval[ left_, right_ ] [ { f_, l_, r_ } ] :=
	{ f, Append[ ToList[l], left ], Append[ ToList[r], right ] }

Unprotect[ tau ];
Clear[ tau ];
Format[ tau[i_] ] := SequenceForm[ "T", i ]
Protect[ tau ]

(*  mDconvolveTwoIntervals  *)
mDconvolveTwoIntervals[{Area[ar1_], a_, a_}, {Area[ar2_], b_, b_}, t_, _, T_] :=
	{{Area[ar1 ar2], a+b, a+b}} 
mDconvolveTwoIntervals[{Area[ar_], a_, a_}, {f_, b_, c_}, t_, _, T_] :=
	{{ar ((f /. t -> t-a) /. T -> 0), b+a, c+a}}
mDconvolveTwoIntervals[{f_, a_, b_}, {Area[ar_], c_, c_}, t_, _, T_] :=
	{{ar ((f /. t -> t-c) /. T -> 0), a+c, b+c}}
mDconvolveTwoIntervals[{ar_, a_, a_}, {f_, b_, c_}, n_, Discrete, T_] :=
	{{ar ((f /. n -> n-a) /. T -> 0), b+a, c+a}}
mDconvolveTwoIntervals[{f_, b_, c_}, {ar_, a_, a_}, n_, Discrete, T_] :=
	{{ar ((f /. n -> n-a) /. T -> 0), b+a, c+a}}

mDconvolveTwoIntervals[{f_, t1_, t2_}, {g_, t3_, t4_}, t_, time_, T_] :=
	mDconvolveTwoIntervals[{g, t3, t4}, {f, t1, t2}, t, time, T] /;
	outOfOrder[t1, t2, t3, t4]

mDconvolveTwoIntervals[{f_, t1_, t2_}, {g_, t3_, t4_}, t_, Continuous, T_] := 
	convolveTwoContinuousIntervals[ f g, t, t1, t2, t3, t4, T ]

mDconvolveTwoIntervals[{f_, n1_, n2_}, {g_, n3_, n4_}, n_, Discrete, T_] := 
	convolveTwoDiscreteIntervals[ f g, n1, n2, n3, n4, T ] 


(*  E N D     P A C K A G E  *)

End[]
EndPackage[]


(*  E N D I N G     M E S S A G E  *)

Print["Multidimensional plotting routines for transforms are loaded."]
Null
